! TXDPF -- Tcode7 Dead Procedure Filter
! Nils M Holm, 1999,2000,2002,2019
! See the file LICENSE for conditions of use.
!
! USAGE: TXDPF [-T tmpdir] <INFILE >OUTFILE

module txdpf(t3x, tcode);

object	t[t3x];

const	BUFLEN		= 1024;		! I/O buffer size
const	MAXLABEL	= 4096;		! max. label ID
const	MAXREF		= 8192;		! max. number of references / module
const	MAXIPROC	= 1024;		! max. number of interface procs

! procedure reference flags
const	DEAD	= -1,	! label unused
	LEAF	= -2;	! label is a leaf (used but has no callees)

var	Inbuf::BUFLEN, Ilen, Iptr;	! input buffer, length, ptr
var	Outbuf::BUFLEN, Optr;		! output buffer, ptr
var	Infile, Outfile;		! input file, output file
var	Procs[MAXLABEL],		! procedure call trees
	Uflag::MAXLABEL;		! procedure 'called' flags
var	Iprocs::MAXIPROC;		! interface procedure usage map
var	CRefs;				! global data referenes to code
var	Entrypoint;			! label of initial entry point
var	Skip;				! suspend output
var	Tmpname::256;			! temporary file name
var	Addr;				! Input position
var	Pass;				! Pass number

! reference (call) structure
struct	REF =	Rdest,	! label of called procedure
		Rnext;	! next REF

var	Refs[REF*MAXREF], Rtop;		! references, top ptr

length(s) return t.memscan(s, 0, 32767);

report(s) t.write(T3X.SYSERR, s, length(s));

reptnl() do var b::3;
	report(t.newline(b));
end

! convert number to padded string
var	ntoa_buf::64;
ntoa(v, w) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
        ntoa_buf::63:= 0;
        i := 62;
        while (v \/ i = 62) do
                ntoa_buf::i := v mod 10 + '0'; i := i-1;
                v := v/10;
		w := w-1;
        end
        if (g) do
                ntoa_buf::i := '-'; i := i-1;
		w := w-1;
        end
	while (w>0) do
		ntoa_buf::i := '\s'; i := i-1;
		w := w-1;
	end
        return @ntoa_buf::(i+1);
end

error(m) do
	report("TXDPF(");
	report(ntoa(Pass, 0));
	report("): ");
	report(ntoa(Addr, 0));
	report(": ");
	report(m);
	reptnl();
	halt 1;
end

rdch() do
	if (Iptr >= Ilen) do
		if (Ilen < 0) error("unexpected EOF");
		Ilen := t.read(Infile, Inbuf, BUFLEN);
		if (Ilen < 1) do
			Ilen := -1;
			return -1;
		end
		Iptr := 0;
	end
	Iptr := Iptr+1;
	Addr := Addr+1;
	return Inbuf::(Iptr-1);
end

flush() do
	if (Optr /\ t.write(Outfile, Outbuf, Optr) \= Optr)
		error("file write error");
	Optr := 0;
end

wrch(c) do
	if (Skip) return 0;
	if (Optr >= BUFLEN) flush();
	Outbuf::Optr := c;
	Optr := Optr+1;
end

rdop() do var a;
	a := rdch();
	a := rdch()<<8 | a;
	return a;
end

wrop(a) do
	wrch(a);
	wrch(a>>8);
end

! advance to next Tcode instruction
next(op) do var a, i;
	wrch(op);
	if (op & 128) do
		a := rdop();
		wrop(a);
		if (op >= tcode.IINIT) do
			a := rdop();
			wrop(a);
		end
		if (	op = tcode.ISTR \/ op = tcode.ILIB \/
			op = tcode.IPUB \/ op = tcode.IEXT \/
			op = tcode.IGSYM \/ op = tcode.ILSYM \/
			op = tcode.IISYM \/ op = tcode.IIPROC \/
			op = tcode.IIREF
		) do
			for (i=0, a) wrch(rdch());
		end
	end
	return rdch();
end

! Add new reference (a) and link into list pointed to by p
newref(p, a) do var n;
	if (Rtop+REF >= REF*MAXREF) error("out of space");
	n := @Refs[Rtop];
	Rtop := Rtop+REF;
	n[Rdest] := a;
	n[Rnext] := p[0];
	p[0] := n;
	return n;
end

! Add reference, if not already contained in p[0]
addref(p, a) do var q;
	q := p[0];
	while (q \= DEAD /\ q \= LEAF) do
		if (q[Rdest] = a) return;
		q := q[Rnext];
	end
	newref(p, a);
end

! Phase 1: build call tree
phase1() do var op, a, plab, prev, v;
	Pass := 1;
	Addr := 0;
	Infile := T3X.SYSIN;
	Outfile := t.open(Tmpname, T3X.OWRITE);
	if (Outfile < 0) error("could not create tmp file");
	plab := -1;	! label of current procedure
	op := rdch();
	prev := -1;	! label created in previous iteration
	while (op >= 0) do
		if (prev \= -1 /\ (op = tcode.IHDR \/ op = tcode.IMHDR)) do
			plab := prev;
			if (plab >= MAXLABEL)
				error("procedure label too big");
			Procs[plab] := LEAF;
		end
		prev := -1;
		ie (	op = tcode.ICALL \/ op = tcode.ILDLAB \/
			op = tcode.ICREF \/ op = tcode.ICLAB
		) do
			wrch(op);
			a := rdop();
			if (a >= MAXLABEL) error("label too big");
			wrop(a);
			ie (op = tcode.ICLAB) do
				ie (a = Entrypoint)	! IEP has no HDR
					plab := a;
				else
					prev := a;
			end
			else ie (op = tcode.ICREF /\ plab = -1) do
				! CREF not in a procedure
				addref(@CRefs, a);
			end
			else do
				if (plab >= MAXLABEL)
					error("procedure label too big");
				! CALL and LDLAB make no sense when not
				!  in a procedure
				if (plab = -1)
					error("illegal reference");
				addref(@Procs[plab], a);
			end
			op := rdch();
		end
		else ie (op = tcode.IICALL) do
			wrch(op);
			a := rdop();
			wrop(a);
			if (a >= MAXIPROC) error("interface label too big");
			Iprocs::a := 1;
			op := rdch();
		end
		else ie (op = tcode.IINIT) do
			wrch(op);
			a := rdop();
			wrop(a);
			v := a;
			if (v \= 7) error("unsupported Tcode version");
			a := rdop();
			wrop(a);
			if (Entrypoint \= DEAD)
				error("multiple entrypoints");
			Entrypoint := a;
			op := rdch();
		end
		else ie (op = tcode.IEND \/ op = tcode.IENDM) do
			plab := -1;
			op := next(op);
		end
		else do
			op := next(op);
		end
	end
	flush();
	t.close(Outfile);
end

! mark all procedures that may be called by p
mark(p) do var q, r;
	if (Uflag::p) return;
	Uflag::p := 1;
	q := Procs[p];
	while (q \= DEAD /\ q \= LEAF) do
		mark(q[Rdest]);
		q := q[Rnext];
	end
end

! Phase 2: mark reachable procedures
phase2() do var i, p;
	Pass := 2;
	Addr := 0;
	for (i=0, MAXLABEL) Uflag::i := 0;
	if (Entrypoint \= DEAD) mark(Entrypoint);
	p := CRefs;
	while (p \= DEAD /\ p \= LEAF) do
		mark(p[Rdest]);
		p := p[Rnext];
	end
end

! Phase 3: copy (saved) input to output, skipping unused procedures
phase3() do var op, o2, a, i, s;
	Pass := 3;
	Addr := 0;
	Ilen := 0;
	Iptr := 0;
	Infile := t.open(Tmpname, T3X.OREAD);
	if (Infile < 0) error("could not reopen tmp file");
	Outfile := T3X.SYSOUT;
	op := rdch();
	while (op >= 0) do
		ie (op = tcode.ICLAB) do
			a := rdop();
			op := rdch();
			if (op = tcode.IHDR \/ op = tcode.IMHDR)
				! Found a procedure;
				! if it is unused, suspend output.
				Skip := \Uflag::a;
			wrch(tcode.ICLAB);
			wrop(a);
			o2 := next(op);
			if (op = tcode.IEND \/ op = tcode.IENDM)
				! Found the end of a procedure
				! Resume output.
				Skip := 0;
			op := o2;
		end
		else ie (op = tcode.IIPROC) do
			a := rdop();
			s := Skip;
			Skip := \Iprocs::a;
			wrch(op);
			wrop(a);
			a := rdop();
			wrop(a);
			for (i=0, a) wrch(rdch());
			op := rdch();
			if (op = tcode.ICMAP) op := next(op);
			Skip := s;
		end
		else ie (op = tcode.IEND \/ op = tcode.IENDM) do
			! Redundant, because END/ENDM is always
			! preceded by a label;
			! however: better safe than sorry!
			Skip := 0;
			op := next(op);
		end
		else do
			op := next(op);
		end
	end
	t.close(Infile);
	flush();
	t.remove(Tmpname);
end

dumprefs() do var i, p;
	for (i=0, MAXLABEL) do
		if (Procs[i] \= DEAD) do
			report(Uflag::i-> "*": "\s");
			report(ntoa(i,5));
			report(": ");
			p := Procs[i];
			while (p \= DEAD /\ p \= LEAF) do
				report(ntoa(p[Rdest],0));
				if (p[Rnext] \= DEAD /\ p[Rnext] \= LEAF)
					report(",");
				p := p[Rnext];
			end
			reptnl();
		end
	end
	p := CRefs;
	report(" CREFs: ");
	while (p \= -1) do
		report(ntoa(p[Rdest],0));
		if (p[Rnext] \= DEAD /\ p[Rnext] \= LEAF)
			report(",");
		p := p[Rnext];
	end
	reptnl();
	report("  INIT: ");
	report(ntoa(Entrypoint,0));
	reptnl();
end

init() do var i;
	ie (t.getarg(1, Tmpname, 256) > 1) do
		ie (Tmpname::0 = '-' /\ Tmpname::1 = 'T') do
			ie (Tmpname::2)
				t.memcopy(Tmpname, @Tmpname::2,
					length(@Tmpname::1));
			else if (t.getarg(2, Tmpname, 256) < 2)
				Tmpname::0 := 0;
		end
		else do
			Tmpname::0 := 0;
		end
	end
	else do
		Tmpname::0 := 0;
	end
	if (\Tmpname::0 /\ t.getenv("TMP", Tmpname, 256) < 0) do
		Tmpname::0 := '.';
		Tmpname::1 := 0;
	end
	i := length(Tmpname);
	if (i >= 245) error("TMP path too long");
	t.memcopy(@Tmpname::i, "/txdpf.tmp", 13);
	Ilen := 0;
	Iptr := 0;
	Optr := 0;
	Addr := 0;
	Infile := 0;
	Outfile := 1;
	for (i=0, MAXLABEL) Procs[i] := DEAD;
	Crefs := DEAD;
	Entrypoint := DEAD;
	Rtop := 0;
	Skip := 0;
	t.memfill(Iprocs, 0, MAXIPROC);
end

do
	init();
	phase1();
	phase2();
	phase3();
	! dumprefs();
end
