! TXLD -- Load Tcode Programs
! Nils M Holm, 1999,2000,2002,2019
! See the file LICENSE for conditions of use.
!
! USAGE: CAT MODULE ... | TXLD >OUTFILE

module txld(t3x, tcode);

object	t[t3x];

const	BUFSIZE	=	1024;	! I/O buffer size
const	TEXTLEN	=	256;	! max. symbol size
const	PUBSIZE	=	6144;	! public symbol table size (words)
const	EXTSIZE	=	2048;	! external symbol map size (words)
const	INFSIZE	=	2048;	! interface symbol table size (words)
const	REFSIZE =	1024;	! interface map size (words)
const	MAXMODS	=	256;	! max. number of input modules
const	LIBMAX =	256;	! space for external module names

! public symbol structure;
! symbols are arranged in a binary tree
struct	PSYM = P_left, P_right, P_lab;
const	P_name = PSYM;	! name is appended to above struct

! external/public mappig structure
struct	XMAP = X_id, X_lab;

var	Buffer::BUFSIZE, Cp, Ep;	! input buffer, char ptr, end ptr
var	Obuffer::BUFSIZE, Obp;		! output buffer, free ptr
var	Symbols;			! ptr to symbol table (Pubtbl)
var	Isymbols;			! ptr to Interface sym. table (Inftbl)
var	Pubtbl[PUBSIZE], Ptop;		! public symbol table, top ptr
var	Exttbl[EXTSIZE], Etop;		! ext. symbol table, top ptr
var	Inftbl[EXTSIZE], Itop;		! interface symbol table, top ptr
var	Reftbl[REFSIZE], Rtop;		! ref. symbol table, top ptr
var	Modlist[MAXMODS], Mtop;		! Module list, top ptr
var	Liblist[LIBMAX], Ltop;		! Library list, top ptr
var	Addr;				! curr. input address
var	K64k;				! 65536 constant
var	Base, Bmax;			! base address, max address+1
var	Ibase, Imax;			! interface base, interface max
var	Errflag;			! Error flag

init() do
	Ep := 1;
	Cp := 1;
	Obp := 0;
	Addr := 0;
	Base := 0;
	Bmax := 0;
	Ibase := 0;
	Imax := 0;
	Ptop := 0;
	Etop := 0;
	Mtop := 0;
	Ltop := 0;
	Itop := 0;
	Rtop := 0;
	Symbols := 0;
	Isymbols := 0;
	K64k := 16384;
	K64k := K64k .* 4;	! 32-bit hack
	Errflag := 0;
end

length(s) return t.memscan(s, 0, 32767);

report(s) t.write(T3X.SYSERR, s, length(s));

reptnl() do var b::3;
	report(t.newline(b));
end

! convert number to string
var	ntoa_buf::32;
ntoa(v) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
        ntoa_buf::31:= 0;
        i := 30;
        while (v \/ i = 30) do
                ntoa_buf::i := v mod 10 + '0'; i := i-1;
                v := v/10;
        end
        if (g) do
                ntoa_buf::i := '-'; i := i-1;
        end
        return @ntoa_buf::(i+1);
end

rdch() do
	if (Cp >= Ep) do
		Ep := t.read(T3X.SYSIN, Buffer, BUFSIZE);
		Cp := 0;
		if (Ep < 0) Ep := 0;
		if (\Ep) return %1;
	end
	Cp := Cp+1;
	Addr := Addr+1;
	return Buffer::(Cp-1);
end

flush() do
	t.write(T3X.SYSOUT, Obuffer, Obp);
	Obp := 0;
end

wrch(n) do
	if (Obp >= BUFSIZE) flush();
	Obuffer::Obp := n;
	Obp := Obp+1;
end

! sign-extend word on non-16-bit machines
extend(x) return x > 32767-> x-K64k: x;

fatal(m) do
	report("TXLD: ");
	report(ntoa(Addr));
	report(": ");
	report(m);
	reptnl();
	halt 1;
end

error(m, n) do var o;
	report("TXLD: ");
	report(n);
	report(": ");
	report(m);
	reptnl();
	Errflag := 1;
end

! find symbol and parent entry (pp)
findsym(name, pp) do var k, y, r, parent;
	y := Symbols;
	parent := 0;
	if (pp) pp[0] := 0;
	k := length(name);
	while (y) do
		r := t.memcomp(name, @y[P_name], k+1);
		if (\r) do
			if (pp /\ parent)
				pp[0] := @parent[r<0-> P_left: P_right];
			return y;
		end
		parent := y;
		y := r<0-> y[P_left] : y[P_right];
	end
	if (pp /\ parent) pp[0] := @parent[r<0-> P_left: P_right];
	return 0;
end

addsym(name, lab) do var k, y, new, pp;
	y := findsym(name, @pp);
	if (y) do
		error("duplicate symbol", name);
		return y;
	end
	k := length(name);
	if ((Ptop + k + 2) / 2 + PSYM >= PUBSIZE)
		fatal("symbol table overflow");
	new := @Pubtbl[Ptop];
	Ptop := Ptop + (k+2)/2 + PSYM;
	new[P_lab] := lab;
	new[P_left] := 0;
	new[P_right] := 0;
	t.memcopy(@new[P_name], name, k+1);
	ie (pp) do
		pp[0] := new;
	end
	else do
		Symbols := new;
	end
	return new;
end

! associate an external label with a public label
addref(name, id) do var y, a, new;
	y := findsym(name, 0);
	ie (\y) do
		error("unresolved external reference", name);
		a := -1;
	end
	else do
		a := y[P_lab];
	end
	if (Etop + XMAP >= EXTSIZE)
		fatal("link map overflow");
	new := @Exttbl[Etop];
	Etop := Etop+XMAP;
	new[X_id] := id;
	new[X_lab] := a;
	return a;
end

lookup(id) do var i;
	for (i=0, Etop, XMAP) do
		if (Exttbl[i+X_id] = id)
			return Exttbl[i+X_lab];
	end
	error("unresolved external label", ntoa(id));
end

! find interface symbol and parent (pp)
findisym(name, pp) do var k, y, r, parent;
	y := Isymbols;
	parent := 0;
	if (pp) pp[0] := 0;
	k := length(name);
	while (y) do
		r := t.memcomp(name, @y[P_name], k+1);
		if (\r) do
			if (pp /\ parent)
				pp[0] := @parent[r<0-> P_left: P_right];
			return y;
		end
		parent := y;
		y := r<0-> y[P_left] : y[P_right];
	end
	if (pp /\ parent) pp[0] := @parent[r<0-> P_left: P_right];
	return 0;
end

addisym(name, lab) do var k, y, new, pp;
	y := findisym(name, @pp);
	if (y) do
		error("duplicate interface symbol", name);
		return y;
	end
	k := length(name);
	if ((Itop + k + 2) / 2 + PSYM >= INFSIZE)
		fatal("interface symbol table overflow");
	new := @Inftbl[Itop];
	Itop := Itop + (k+2)/2 + PSYM;
	new[P_lab] := lab;
	new[P_left] := 0;
	new[P_right] := 0;
	t.memcopy(@new[P_name], name, k+1);
	ie (pp) do
		pp[0] := new;
	end
	else do
		Isymbols := new;
	end
	return new;
end

! associate an external label with a real (public) label
addiref(name, id) do var y, a, new;
	y := findisym(name, 0);
	ie (\y) do
		error("unresolved interface", name);
		a := -1;
	end
	else do
		a := y[P_lab];
	end
	if (Rtop + XMAP >= REFSIZE)
		fatal("interface reference map overflow");
	new := @Reftbl[Rtop];
	Rtop := Rtop+XMAP;
	new[X_id] := id;
	new[X_lab] := a;
	return a;
end

ilookup(id) do var i;
	for (i=0, Rtop, XMAP) do
		if (Reftbl[i+X_id] = id)
			return Reftbl[i+X_lab];
	end
	error("unresolved interface reference", ntoa(id));
end

! memorize init address of module
addmodule(id) do
	if (Mtop >= MAXMODS) fatal("too many modules");
	Modlist[Mtop] := id;
	Mtop := Mtop+1;
end

addlib(name, k) do
	if (Ltop + k > LIBMAX) fatal("too many libraries");
	Liblist::Ltop := k;
	t.memcopy(@Liblist::(Ltop+1), name, k);
	Ltop := Ltop + k+1;
end

readop() do var a;
	a := rdch();
	a := rdch()<<8 | a;
	return extend(a);
end

writeop(n) do
	wrch(n);
	wrch(n>>8);
end

gen1(i, o) do
	wrch(i);
	writeop(o);
end

load() do
	var	op, c, a1, a2, s, k, i, n, o2;
	var	proc, sbuf::TEXTLEN;

	addlib("t3x", 4);
	while (1) do
		proc := 0;	! 0 = pass thru, 1 = process
		op := rdch();
		if (\Ep) leave;
		if (op & 127 >= tcode.IENDOFSET) fatal("illegal opcode");
		! get number of args
		c := op & 128-> op >= tcode.IINIT-> 2: 1: 0;
		! get string flag
		s := op = tcode.ISTR \/ op = tcode.ILIB \/
			op = tcode.IPUB \/ op = tcode.IEXT \/
			op = tcode.IGSYM \/ op = tcode.ILSYM \/
			op = tcode.IISYM \/ op = tcode.IIPROC \/
			op = tcode.IIREF; 
		if (c) a1 := readop();
		if (c > 1) a2 := readop();
		if (op = tcode.IINIT) do
			if (a1 \= 7) fatal("unsupported Tcode version");
			! discard CALL,HALT
			for (i=0, 2) do
				rdch();
				readop();
			end
			Base := Bmax;
			Ibase := Imax;
			! clear external reference tables
			Etop := 0;
			Rtop := 0;
		end
		ie (op = tcode.ICLAB \/ op = tcode.IDLAB \/
			op = tcode.ICREF \/ op = tcode.IDREF \/
			op = tcode.IINCG \/ op = tcode.ILDG \/
			op = tcode.ILDGV \/ op = tcode.ILDLAB \/
			op = tcode.ISAVG \/ op = tcode.ICALL \/
			op = tcode.IBRF \/ op = tcode.IBRT \/
			op = tcode.INBRF \/ op = tcode.INBRT \/
			op = tcode.IJUMP \/ op = tcode.IUNEXT \/
			op = tcode.IDNEXT \/ op = tcode.IPUB \/
			op = tcode.IINIT
		) do
			! adjust address (relocate)
			ie (op = tcode.IINIT) do
				a2 := a2 + Base;
				if (a2 >= Bmax) Bmax := a2+1;
			end
			else do
				a1 := a1 + Base;
				if (a1 >= Bmax) Bmax := a1+1;
			end
		end
		else ie (op = tcode.IIPROC) do
			! adjust address (relocate)
			a1 := a1 + Ibase;
			if (a1 >= Imax) Imax := a1;
		end
		else ie (op = tcode.ICALX) do
			i := lookup(a1);
			if (i \= -1) do
				op := tcode.ICALL;
				a1 := i;
			end
		end
		else if (op = tcode.IICALX) do
			i := ilookup(a1);
			if (i \= -1) do
				op := tcode.IICALL;
				a1 := i;
			end
		end
		if (	op = tcode.IEXT \/ op = tcode.IPUB \/
			op = tcode.IIREF \/ op = tcode.ILIB
		)
			proc := 1;
		if (op = tcode.IINIT) do
			addmodule(a2);
			proc := 1;		! swallow all but first INIT
			if (\Base) do;		! redirect first INIT to CLAB 0
				wrch(tcode.IINIT);
				writeop(a1);
				writeop(0);
			end
		end
		if (\proc) do
			wrch(op);
			if (c) writeop(a1);
			if (c > 1) writeop(a2);
		end
		if (s) do
			k := (op = tcode.ISTR \/ op = tcode.ILIB)-> a1: a2;
			for (i=0, k) do
				n := rdch();
				sbuf::i := n;
				if (\proc) wrch(n);
			end
			sbuf::i := 0;
			if (op = tcode.ILIB /\ t.memcomp(sbuf, "t3x", 4) \= 0)
				addlib(sbuf, a1+1);
		end
		if (Base = 0 /\ op = tcode.IINIT)
			gen1(tcode.IJUMP, 0);	! jump to main procedure
		if (op = tcode.IPUB) addsym(sbuf, a1);
		if (op = tcode.IIPROC) addisym(sbuf, a1);
		if (op = tcode.IEXT) i := addref(sbuf, a1) = -1;
		if (op = tcode.IIREF) i := addiref(sbuf, a1) = -1;
		if (i = -1) do
			wrch(op);
			writeop(a1);
			writeop(k);
			for (i=0, k) wrch(sbuf::i);
		end
	end
	! the new main procedure begins here
	gen1(tcode.ICLAB, 0);
	! emit an init call for each module
	for (i=0, Mtop) if (Modlist[i]) do
		gen1(tcode.ICALL, Modlist[i]);
		gen1(tcode.ICLEAN, 0);
		wrch(tcode.IPOP);
	end
	! in case the main program returns
	gen1(tcode.IHALT, 0);
end

wrlibdir() do var ld, i, nl::3, k;
	if (Ltop = 0) return;
	ld := t.open("LIBDIR", T3X.OWRITE);
	if (ld = -1) fatal("cannot create LIBDIR file");
	i := 0;
	t.newline(nl);
	k := nl::1-> 2: 1;
	while (i < Ltop) do
		t.write(ld, @Liblist::(i+1), Liblist::i - 1);
		t.write(ld, nl, k);
		i := i + Liblist::i + 1;
	end
	t.close(ld);
end

do
	init();
	load();
	flush();
	wrlibdir();
	if (Errflag) halt 1;
end
