! TXOPT -- A Tcode Optimizer.
! Nils M Holm, 1999,2000,2002,2019
! See the file LICENSE for conditions of use.
!
! USAGE: TXOPT <INFILE >OUTFILE

module txopt(t3x, tcode);

object	t[t3x];

const	BUFLEN		= 1024;	! I/O buffer size
const	CODESIZE	= 8192;	! max. code size per procedure
const	NLABELS		= 4096;	! max. labels per procedure
const	STACKLEN	= 64;	! stack size for folding const exprs
const	IBUFLEN	= 512;		! instruction buffer size for tree opts

! Tcode instruction node
struct	TCI = tc_op, tc_arg, tc_left, tc_right;

var	Code::CODESIZE;			! code buffer
var	Ctop;				! code size
var	Labels[NLABELS], Lbase;		! Labels, base label
var	Stack[STACKLEN], Sp;		! folding stack, ptr
var	Inbuf::BUFLEN, Cp, Ep;		! input buffer, char ptr, end ptr
var	Comtab;				! commutative op table
var	Ibuffer[IBUFLEN*TCI], Ip0, Ip;	! Instr buffer, code ptr, stack ptr
var	Checkmagic;			! check magic flag, 1=check
var	K64k;				! 65536 constant
var	Addr;				! Input position

! Comtab structure
struct CT = CPREC, COPER, CREPL;

init() do
	Cp := 0;
	Ep := 0;
	Checkmagic := 1;
	K64k := 16384;	! avoid constant expression folding
	K64k := K64k * 4;
	Addr := 0;
	Code::(CODESIZE-1) := tcode.ILDG;
	Comtab := [
	! P  instruction	replacement	(P = precedence)
	[ 1, tcode.INORM,	0		],
	[ 1, tcode.INORMB,	tcode.INORMB	],
	[ 1, tcode.IDEREF,	0		],
	[ 1, tcode.IDREFB,	0		],
	[ 2, tcode.IMUL,	tcode.IMUL	],
	[ 2, tcode.IUMUL,	tcode.IUMUL	],
	[ 2, tcode.IDIV,	0		],
	[ 2, tcode.IUDIV,	0		],
	[ 2, tcode.IMOD,	0		],
	[ 3, tcode.IADD,	tcode.IADD	],
	[ 3, tcode.ISUB,	0		],
	[ 4, tcode.IBAND,	tcode.IBAND	],
	[ 4, tcode.IBOR,	tcode.IBOR	],
	[ 4, tcode.IBXOR,	tcode.IBXOR	],
	[ 4, tcode.IBSHL,	0		],
	[ 4, tcode.IBSHR,	0		],
	[ 5, tcode.IEQU,	tcode.IEQU	],
	[ 5, tcode.INEQU,	tcode.INEQU	],
	[ 6, tcode.ILESS,	tcode.IGRTR	],
	[ 6, tcode.IGRTR,	tcode.ILESS	],
	[ 6, tcode.ILTEQ,	tcode.IGTEQ	],
	[ 6, tcode.IGTEQ,	tcode.ILTEQ	],
	[ 6, tcode.IULESS,	tcode.IUGRTR	],
	[ 6, tcode.IUGRTR,	tcode.IULESS	],
	[ 6, tcode.IULTEQ,	tcode.IUGTEQ	],
	[ 6, tcode.IUGTEQ,	tcode.IULTEQ	],
	[ 0, 0,			0		],
	];
end

length(s) return t.memscan(s, 0, 32767);

report(s) t.write(T3X.SYSERR, s, length(s));

reptnl() do var b::3;
	report(t.newline(b));
end

! convert number to string
var	ntoa_buf::64;
ntoa(v) do var g, i;
	g := 0;
	if (v<0) do
		g := 1;
		v := -v;
	end
        ntoa_buf::63:= 0;
        i := 62;
        while (v \/ i = 62) do
                ntoa_buf::i := v mod 10 + '0'; i := i-1;
                v := v/10;
        end
        if (g) do
                ntoa_buf::i := '-'; i := i-1;
        end
        return @ntoa_buf::(i+1);
end

fatal(m, n) do
	report("TXOPT: ");
	report(ntoa(Addr));
	report(": ");
	report(m);
	if (n) do
		report(": ");
		report(n);
	end
	reptnl();
	report("terminating."); reptnl();
	halt 1;
end

getw(a) do var n;
	n := Code::(a+1) << 8 | Code::a;
	return (n > 32767)-> n-K64k: n;		! non 16-bit hack !
end

putw(n, v) do
	Code::n := v & 255;
	Code::(n+1) := v >> 8;
end

! advance to next Tcode instruction (starting at position p)
next(p) do var i;
	i := Code::p;
	if (i = tcode.IINIT) return p+5;
	p := p+1;
	if (i & 128) do
		p := p+2;
                if (i >= tcode.IINIT) p := p+2;
                if (    i = tcode.ISTR \/ i = tcode.ILIB \/
                        i = tcode.IPUB \/ i = tcode.IEXT \/
                        i = tcode.IGSYM \/ i = tcode.ILSYM \/
                        i = tcode.IISYM \/ i = tcode.IIPROC \/
			i = tcode.IIREF
                )
			p := p + getw(p-2);
	end
	return p;
end

rdch() do
	if (Cp >= Ep) do
		if (Ep = -1) return -1;
		Ep := t.read(T3X.SYSIN, Inbuf, BUFLEN);
		if (Ep < 1) do
			Ep := -1;
			return -1;
		end
		Cp := 0;
	end
	Cp := Cp+1;
	Addr := Addr+1;
	return Inbuf::(Cp-1);
end

readop(p) do var k;
	k := rdch();
	k := rdch()<<8 | k;
	p::0 := k;
	p::1 := k>>8;
	return k;
end

! copy an instruction into memory
! i = first byte of instruction to copy
copyi(i) do var k, j;
	Code::Ctop := i;
	Ctop := Ctop+1;
	if (Ctop >= CODESIZE-6) fatal("procedure too big", 0);
	if (i & 128) do
		k := readop(@Code::Ctop);
		Ctop := Ctop+2;
		! adjust Lbase if necessary
		if (i = tcode.ICLAB /\ k < Lbase) Lbase := k;
		if (i >= tcode.IINIT) do
			if (i = tcode.IINIT /\ k \= 7)
				fatal("unsupported Tcode version",
					ntoa(k));
			j := k;
			k := readop(@Code::Ctop);
			Ctop := Ctop+2;
		end
		if (    i = tcode.ISTR \/ i = tcode.ILIB \/
			i = tcode.IPUB \/ i = tcode.IEXT \/
			i = tcode.IGSYM \/ i = tcode.ILSYM \/
			i = tcode.IISYM \/ i = tcode.IIPROC \/
			i = tcode.IIREF
		) do
			if (Ctop + k >= CODESIZE)
				fatal("procedure too big", 0);
			for (j=0, k) do
				Code::(Ctop+j) := rdch();
			end
			Ctop := Ctop+k;
		end
	end
	return rdch();
end

! read a procedure and its prefix (data declarations) into memory
readproc() do var op, k;
	Ctop := 0;
	Lbase := 32767;	! will be adjusted by copyi()
	op := rdch();
	if (Checkmagic /\ op \= tcode.IINIT)
		fatal("magic match failed", 0);
	Checkmagic := 0;
	! read everything up to beginning of next procedure
	while (op \= tcode.IHDR /\ op \= tcode.IMHDR /\ op \= -1)
		op := copyi(op);
	if (op = -1) return -1;
	! read everything up to end of procedure
	while (op \= tcode.IEND /\ op \= tcode.IENDM /\ op \= -1)
		op := copyi(op);
	if (op = -1) fatal("unexpected end of procedure", 0);
	Code::Ctop := op;
	Ctop := Ctop+1;
end

! 'nop out' a three-byte instruction
nop3(p) do
	Code::p := tcode.IGLUE;
	Code::(p+1) := tcode.IGLUE;
	Code::(p+2) := tcode.IGLUE;
end

! fold multiply instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_mul(p, a1, a2, t1, t2, v1, v2) do
	if (	t1 = tcode.INUM /\ v1 = 0 \/		! x*0, 0*x --> 0
		t2 = tcode.INUM /\ v2 = 0
	) do
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = 1) do		! x*1 --> x
		nop3(a1);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! 1*x --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = %1) do		! x*-1 --> -x
		nop3(a1);
		Code::p := tcode.INEG;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = %1) do		! -1*x --> -x
		nop3(a2);
		Code::p := tcode.INEG;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold divide instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_div(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0/x --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x/0 --> _|_
		fatal("constant divide by zero", 0);
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! x/1 --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = %1) do		! x/-1 --> -x
		nop3(a2);
		Code::p := tcode.INEG;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold modulo instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_mod(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0 mod x --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x mod 0 --> _|_
		fatal("constant divide by zero", 0);
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! x mod 1 --> 0
		nop3(a2);
		Code::p := tcode.IGLUE;
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold add instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_add(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0+x --> x
		nop3(a1);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x+0 --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold subtract instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_sub(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0-x --> -x
		nop3(a1);
		Code::p := tcode.INEG;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x-0 --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold binary AND instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_band(p, a1, a2, t1, t2, v1, v2) do
	if (	t1 = tcode.INUM /\ v1 = 0 \/
		t2 = tcode.INUM /\ v2 = 0
	) do						! 0&x, x&0 --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = %1) do		! %1&x --> x
		nop3(a1);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = %1) do		! x&%1 --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold binary OR instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_bor(p, a1, a2, t1, t2, v1, v2) do
	if (	t1 = tcode.INUM /\ v1 = %1 \/
		t2 = tcode.INUM /\ v2 =%1 
	) do						! %1|x, x|%1 --> %1
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, %1);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0|x --> x
		nop3(a1);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x|0 --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold binary SHIFT LEFT instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_bshl(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0<<x --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x<<0 --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold binary SHIFT RIGHT instructions - same as SHIFT LEFT, above
fold_bshr(p, a1, a2, t1, t2, v1, v2)
	return fold_bshl(p, a1, a2, t1, t2, v1, v2);
	
! fold binary XOR instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_bxor(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = %1) do		! x^%1 --> ~x
		nop3(a1);
		Code::p := tcode.IBNOT;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = %1) do		! x^%1 --> ~x
		nop3(a2);
		Code::p := tcode.IBNOT;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold NORM instruction (pointer normalization)
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_norm(p, a1, a2, t1, t2, v1, v2) do
	! Fold the address computation of local vector members
	! at constant offsets
	if (t1 = tcode.ILDLV /\ t2 = tcode.INUM) do	! @x[N] --> @(x-N)
		putw(a1+1, v1 - v2);			! for local x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! @x[0] --> @x
		nop3(a2);
		Code::p := tcode.IGLUE;
		return %1;
	end
	return 0;
end

! fold unsigned multiply instructions
! p  = position of instruction in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_umul(p, a1, a2, t1, t2, v1, v2) do
	if (	t1 = tcode.INUM /\ v1 = 0 \/		! x.*0, 0.*x --> 0
		t2 = tcode.INUM /\ v2 = 0
	) do
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t1 = tcode.INUM /\ v1 = 1) do		! x.*1 --> x
		nop3(a1);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		Stack[Sp-1] := a2;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! 1.*x --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! p  = position of instruction in Code array
! p  = position in Code array
! a1 = address of first argument
! a2 = address of second argument
! t1 = type of first argument
! t2 = type of second argument
! v1 = value of first argument
! v2 = value of second argument
fold_udiv(p, a1, a2, t1, t2, v1, v2) do
	if (t1 = tcode.INUM /\ v1 = 0) do		! 0./x --> 0
		nop3(a2);
		Code::a1 := tcode.INUM;
		putw(a1+1, 0);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	if (t2 = tcode.INUM /\ v2 = 0) do		! x./0 --> _|_
		fatal("constant divide by zero", 0);
	end
	if (t2 = tcode.INUM /\ v2 = 1) do		! x./1 --> x
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;
		return %1;
	end
	return 0;
end

! fold binary expressions
! p  = position of instruction in Code array
fold_binary(p) do var op, t1, t2, a1, a2, v1, v2;
	op := Code::p;
	a1 := Stack[Sp-2];
	a2 := Stack[Sp-1];
	t1 := Code::a1;
	t2 := Code::a2;
	v1 := getw(a1+1);
	v2 := getw(a2+1);
	Sp := Sp-1;
	! Fold operations with two constant operands
	! constant OP constant --> constant
	if (t1 = tcode.INUM /\ t2 = tcode.INUM) do
		if (\v2 /\ (op = tcode.IMOD \/ op = tcode.IDIV))
			fatal("constant divide by zero", 0);
		ie (op = tcode.IMUL) v1 := v1 * v2;
		else ie (op = tcode.IDIV) v1 := v1 / v2;
		else ie (op = tcode.IMOD) v1 := v1 mod v2;
		else ie (op = tcode.IADD) v1 := v1 + v2;
		else ie (op = tcode.ISUB) v1 := v1 - v2;
		else ie (op = tcode.IBAND) v1 := v1 & v2;
		else ie (op = tcode.IBOR) v1 := v1 | v2;
		else ie (op = tcode.IBXOR) v1 := v1 ^ v2;
		else ie (op = tcode.IBSHL) v1 := v1 << v2;
		else ie (op = tcode.IBSHR) v1 := v1 >> v2;
		else ie (op = tcode.IEQU) v1 := v1 = v2;
		else ie (op = tcode.INEQU) v1 := v1 \= v2;
		else ie (op = tcode.ILESS) v1 := v1 < v2;
		else ie (op = tcode.IGRTR) v1 := v1 > v2;
		else ie (op = tcode.ILTEQ) v1 := v1 <= v2;
		else ie (op = tcode.IGTEQ) v1 := v1 >= v2;
		else ie (op = tcode.IUMUL) v1 := v1 .* v2;
		else ie (op = tcode.IUDIV) v1 := v1 ./ v2;
		else ie (op = tcode.IULESS) v1 := v1 .< v2;
		else ie (op = tcode.IUGRTR) v1 := v1 .> v2;
		else ie (op = tcode.IULTEQ) v1 := v1 .<= v2;
		else if (op = tcode.IUGTEQ) v1 := v1 .>= v2;
		putw(a1+1, v1);
		nop3(a2);
		Code::p := tcode.IGLUE;
		Sp := Sp-1;	!XXX ???
		return;
	end
	if (op = tcode.INORM /\ fold_norm(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IMUL /\ fold_mul(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IDIV /\ fold_div(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IMOD /\ fold_mod(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IADD /\ fold_add(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.ISUB /\ fold_sub(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IBAND /\ fold_band(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IBOR /\ fold_bor(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IBSHL /\ fold_bshl(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IBSHR /\ fold_bshl(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IBXOR /\ fold_bxor(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IUMUL /\ fold_umul(p, a1, a2, t1, t2, v1, v2)) return;
	if (op = tcode.IUDIV /\ fold_udiv(p, a1, a2, t1, t2, v1, v2)) return;
	! otherwise, make sure the result is non-constant
	Stack[Sp-1] := CODESIZE-1;	! tcode.ILDG
end

! fold increments
! (make advantage of the INCx instructions)
! p     = position of SAVx
! start = beginnig of potential increment (LDx a NUM y ADD SAVx)
fold_inc(p, start) do		! x := x+N --> inc(x,N)
	var	op, a;		! x := x-N --> inc(x,-N)
	var	q, q2;

	op := Code::p;
	if (	Code::start = tcode.ILDL /\ op = tcode.ISAVL \/
		Code::start = tcode.ILDG /\ op = tcode.ISAVG \/
		Code::start = tcode.ILDI /\ op = tcode.ISAVI
	) do
		! make sure that LD and SAV reference the same address
		! and the increment is constant
		a := getw(start+1);
		if (a \= getw(p+1)) return;
		q := next(start);
		if (Code::q \= tcode.INUM) return;
		q2 := next(q);
		if (Code::q2 \= tcode.IADD /\ Code::q2 \= tcode.ISUB) return;
		! replace LDx a NUM y ADD SAVx with INCx a y
		Code::start := Code::start = tcode.ILDL-> tcode.IINCL:
				Code::start = tcode.ILDI-> tcode.IINCI:
					tcode.IINCG;
		putw(start+2+1, Code::q2 = tcode.ISUB-> -getw(start+2+2):
			getw(start+2+2));
		nop3(start+2*2+1);
		nop3(p);
	end
end

! fold all expressions in the current procedure
foldexpr() do
	var	i, op, a, v;
	var	state;
	var	last[3];

	Sp := 0;
	i := 0;
	last[0] := 0;	! queue, used to match increments
	last[1] := 0;
	last[2] := 0;
	while (i < Ctop) do
		op := Code::i;
		ie (	op = tcode.ILDG \/ op = tcode.ILDGV \/
			op = tcode.ILDL \/ op = tcode.ILDLV \/
			op = tcode.ILDLAB \/ op = tcode.INUM
		) do
			if (Sp+1 >= STACKLEN)
				fatal("FOLDEXPR: stack overflow", 0);
			Stack[Sp] := i;
			Sp := Sp+1;
		end
		else ie (Sp /\ (op = tcode.INEG \/ op = tcode.ILNOT \/
			op = tcode.IBNOT)
		) do
			! unary operations, fold if argument is constant
			ie (Code::(Stack[Sp-1]) = tcode.INUM) do
				a := Stack[Sp-1];
				v := getw(a+1);
				ie (op = tcode.INEG) v := -v;
				else ie (op = tcode.ILNOT) v := \v;
				else v := ~v;
				putw(a+1, v);
				Code::i := tcode.IGLUE;
			end
			else do
				! make sure result is non-constant
				Stack[Sp-1] := CODESIZE-1;
			end
		end
		else ie (Sp > 1 /\
			(op = tcode.IMUL \/
			op = tcode.IDIV \/ op = tcode.IMOD \/
			op = tcode.IADD \/ op = tcode.ISUB \/
			op = tcode.IBAND \/ op = tcode.IBOR \/
			op = tcode.IBXOR \/ op = tcode.IBSHL \/
			op = tcode.IBSHR \/
			op = tcode.IEQU \/ op = tcode.INEQU \/
			op = tcode.ILESS \/ op = tcode.IGRTR \/
			op = tcode.ILTEQ \/ op = tcode.IGTEQ \/
			op = tcode.IUMUL \/ op = tcode.IUDIV \/
			op = tcode.IULESS \/
			op = tcode.IUGRTR \/ op = tcode.IULTEQ \/
			op = tcode.IUGRTR)
		) do
			fold_binary(i);
		end
		else ie (op = tcode.IDLAB \/ op = tcode.IVEC \/
			op = tcode.IDATA \/ op = tcode.ICREF \/
			op = tcode.IDREF \/ op = tcode.ILIB \/
			op = tcode.ISTR \/ op = tcode.IGLUE \/
			op = tcode.ILINE \/ op = tcode.IGSYM \/
			op = tcode.ILSYM \/ op = tcode.IISYM \/
			op = tcode.IIPROC \/ op = tcode.IIREF
		) do
			! ignore
		end
		else do
			Sp := 0;
			! SAVx may be the final instruction of an increment
			if (	(op = tcode.ISAVL \/ op = tcode.ISAVG \/
				 op = tcode.ISAVI) /\
				last[0]
			)
				fold_inc(i, last[0]);
		end
		last[0] := last[1];
		last[1] := last[2];
		last[2] := i;
		i := next(i);
	end
end

! fold all conditions in the current procedure
foldcond() do var	i, v, op, last;
	i := 0;
	last := 0;	! last non-NOP instruction
	while (i < Ctop) do
		op := Code::i;
		ie (	(op = tcode.IBRF \/ op = tcode.IBRT) /\
			Code::last = tcode.INUM
		) do
			v := getw(last+1);
			! NUM T BRT L --> JUMP L
			! NUM F BRT L --> NOP
			ie (op = tcode.IBRT) do
				nop3(last);
				ie (v)
					Code::i := tcode.IJUMP;
				else
					nop3(i);
			end
			! NUM T BRF L --> NOP
			! NUM F BRF L --> JUMP L
			else do ! BRF
				nop3(last);
				ie (v)
					nop3(i);
				else
					Code::i := tcode.IJUMP;
			end
		end
		else if ((op = tcode.IBRF \/ op = tcode.IBRT) /\
			Code::last = tcode.ILNOT
		) do
			! NOT BRT  --> BRF
			! NOT BRF  --> BRT
			! NOT NBRT --> NBRF
			! NOT NBRF --> NBRT
			Code::i := op = tcode.IBRT-> tcode.IBRF:
				op = tcode.IBRF-> tcode.IBRT:
				op = tcode.INBRT-> tcode.INBRF:
				tcode.INBRT;
			Code::last := tcode.IGLUE;
		end
		if (Code::i \= tcode.IGLUE) last := i;
		i := next(i);
	end
end

! remove all NOPs from the current procedure
condense() do var i, j, k, m;
	i := 0;
	j := 0;
	while (i < Ctop) do
		k := next(i) - i;
		if (Code::i \= tcode.IGLUE) do
			t.memcopy(@Code::j, @Code::i, k);
			j := j+k;
		end
		i := i+k;
	end
	k := Ctop;
	Ctop := j;
	return k-Ctop;
end

isload(x) return x & 127 >= tcode.ILDG & 127 /\ x & 127 <= tcode.ISELF;

isbinop(x) return x >= tcode.IMUL /\ x <= tcode.IUGTEQ \/
		x >= tcode.IDEREF /\ x <= tcode.INORMB;

isexpr(x) return
	x = tcode.INEG \/
	x = tcode.ILNOT \/
	x = tcode.IBNOT \/
	isbinop(x) \/
 	isload(x) \/
	x = tcode.IDUP \/
	x = tcode.ISWAP;

! check if A and B are the same effect-free subscript
! (as in v[x+y] := v[x+y] + 1;)
!          ^^^       ^^^
common(a, b) do var op;
	if (a = %1) return b = %1;
	if (b = %1) return a = %1;
	if (\common(Ibuffer[a+tc_left], Ibuffer[b+tc_left])) return 0;
	if (\common(Ibuffer[a+tc_right], Ibuffer[b+tc_right])) return 0;
	op := Ibuffer[a+tc_op];
	if (op \= Ibuffer[b+tc_op]) return 0;
	if (Ibuffer[a+tc_arg] \= Ibuffer[b+tc_arg]) return 0;
	! A procedure call in a subscript may have a side-effect!
	if (	op = tcode.ICALL \/ op = tcode.ICALR \/
		op = tcode.ICALX \/
		op = tcode.IICALL \/ op = tcode.IICALX
	)
		return 0;
	return %1;
end

! elimininate one commom subscript (EXPR) from an assignment of the form
!
! VAR[EXPR] := VAR[EXPR] op EXPR2 
! or
! VAR::EXPR := VAR::EXPR op EXPR2
!
! I.e. convert
!
!      this tree      to       that tree
!
!        STORE                   STORE
! a0    /     \           a0    /     \
!   NORM       \ a1         NORM       \ a1
!  /    \      OP          /    \      OP
! VAR  EXPR   /  \        VAR  EXPR   /  \
!            /    \                  /    \
!         DEREF   EXPR2           DEREF   EXPR2
!        /     \                 /     \
!       VAR   EXPR              DUP   NUM 0
!       ^^^   ^^^^              ^^^   ^^^^^

commref(args) do var a0, a1, op, op2, new;
	a0 := args[0];	! tree, lefthand side of :=
	a1 := args[1];	! tree, righthand side of :=
	if (a0 = %1 \/ a1 = %1 \/ Ibuffer[a1+tc_left] = %1) return;
	op := Ibuffer[a0+tc_op];
	if (op \= tcode.INORM /\ op \= tcode.INORMB) return;
	op2 := Ibuffer[Ibuffer[a1+tc_left]+tc_op];
	if (	\((op = tcode.INORM /\ op2 = tcode.IDEREF) \/
		 (op = tcode.INORMB /\ op2 = tcode.IDREFB))
	)
		return;
	if (	\common(Ibuffer[a0+tc_left],
			Ibuffer[Ibuffer[a1+tc_left]+tc_left])
	)
		return;
	if (	\common(Ibuffer[a0+tc_right],
			Ibuffer[Ibuffer[a1+tc_left]+tc_right])
	)
		return;
	! Replace VAR EXPR DEREF on righthand side
	! with DUP NUM 0 DEREF,
	! where DUP duplicates VAR EXPR NORM.
	! This also works with NORMB,DREFB.
	a1 := Ibuffer[a1+tc_left];
	if (a1 = %1 \/ Ibuffer[a1+tc_left] = %1 \/ Ibuffer[a1+tc_right] = %1)
		return;
	new := Ibuffer[a1+tc_left];
	Ibuffer[new+tc_op] := tcode.IDUP;
	Ibuffer[new+tc_left] := %1;
	Ibuffer[new+tc_right] := %1;
	new := Ibuffer[a1+tc_right];
	Ibuffer[new+tc_op] := tcode.INUM;
	Ibuffer[new+tc_arg] := 0;
	Ibuffer[new+tc_left] := %1;
	Ibuffer[new+tc_right] := %1;
end

! Re-order the arguments of commutative operations
! to minimize spilling to the stack. For example,
!	A B C + *
! uses three stack cells, but
!	B C + A *
! uses only two. Of course, this only works because
!	A*B = B*A
! Note that A<B can be replaced with B>A.

reorder(p) do var op, base, left, right, tmp;
	if (p = %1) return;		! NIL
	reorder(Ibuffer[p+tc_left]);
	reorder(Ibuffer[p+tc_right]);
	op := Ibuffer[p+tc_op];
	if (\isbinop(op)) return;
	! Search op in Comtab,
	! IENDOFSET is a resonable limit.
	for (base=0, tcode.IENDOFSET) do
		if (\Comtab[base][CPREC]) return;	! end of table
		if (op = Comtab[base][COPER]) do
			! No replacement? Give up for now.
			! (May also inject SWAP.)
			if (\Comtab[base][CREPL]) return;
			leave;
		end
	end
	! find opcodes of arguments
	op := Ibuffer[Ibuffer[p+tc_left]+tc_op];
	for (left=0, tcode.IENDOFSET)
		if (\Comtab[left][CPREC] \/ op = Comtab[left][COPER])
			leave;
	op := Ibuffer[Ibuffer[p+tc_right]+tc_op];
	for (right=0, tcode.IENDOFSET)
		if (\Comtab[right][CPREC] \/ op = Comtab[right][COPER])
			leave;
	! if no information is available on either argument, give up
	if (\Comtab[right][CPREC] \/ Comtab[left][CPREC]) return;
	! if the right operand has higher precedence (lower value)
	! than the left one, swap them
	if (Comtab[right][CPREC] < Comtab[base][CPREC]) do
		tmp := Ibuffer[p+tc_left];
		Ibuffer[p+tc_left] := Ibuffer[p+tc_right];
		Ibuffer[p+tc_right] := tmp;
		! replace the opcode
		Ibuffer[p+tc_op] := Comtab[base][CREPL];
	end
end

! write back a (modified) tree
writeback(p) do
	if (p = %1) return 0;		! NIL
	writeback(Ibuffer[p+tc_left]);
	writeback(Ibuffer[p+tc_right]);
	! glue node? return.
	if (Ibuffer[p+tc_op] = tcode.IGLUE) return 0;
	Code::Ip0 := Ibuffer[p+tc_op];
	Ip0 := Ip0+1;
	if (Ibuffer[p+tc_op] & 128) do
		Code::Ip0 := Ibuffer[p+tc_arg];
		Code::(Ip0+1) := Ibuffer[p+tc_arg] >> 8;
		Ip0 := Ip0+2;
	end
end

dumptree(p,d) do var i;
	if (p = %1) return 0;
	if (d > 1) for (i=0, d-1) report(".");
	report(ntoa(Ibuffer[p+tc_op]));
	report("\s");
	report(ntoa(Ibuffer[p+tc_arg]));
	reptnl();
	if (Ibuffer[p+tc_left] \= %1) report("L");
	dumptree(Ibuffer[p+tc_left], d+1);
	if (Ibuffer[p+tc_right] \= %1) report("R");
	dumptree(Ibuffer[p+tc_right], d+1);
end

! build tree, perform CSE and term reordering, generate new code
treeopts() do
	var	i, j, n, p, optdone;
	var	args[IBUFLEN], ap;
	var	op;

	i := 0;
	! Ip0 holds the address of the code from which the tree was generated.
	Ip0 := %1;
	optdone := 0;		! optimization successful?
	while (i < Ctop) do
		ie (isexpr(Code::i)) do
			if (Ip0 = %1) do
				Ip0 := i;
				Ip := 0;
				ap := 0;
			end
			op := Code::i;
			! if expr to big, give up
			if (Ip >= IBUFLEN*TCI-TCI) Ip0 := %1;
			! append instruction
			Ibuffer[Ip+tc_op] := op;
			ie (op & 128) do
				Ibuffer[Ip+tc_arg] :=
					Code::(i+2)<<8 | Code::(i+1);
			end
			else do
				Ibuffer[Ip+tc_arg] := 0;
			end
			ie (isload(op)) do
				args[ap] := Ip;
				ap := ap+1;
				! load instructions are leaves
				Ibuffer[Ip+tc_left] := %1;
				Ibuffer[Ip+tc_right] := %1;
			end
			else ie (isbinop(op)) do
				ie (ap < 2) do
					Ip0 := %1;
				end
				else do
					! link arguments to top expr
					Ibuffer[Ip+tc_left] := args[ap-2];
					Ibuffer[Ip+tc_right] := args[ap-1];
					! remove right arg
					ap := ap-1;
					! replace left arg with complete
					! binary operation
					args[ap-1] := Ip;
				end
			end
			else do
				! if there is something in the buffer,
				! it must be a unary operation
				ie (Ip) do
					! link argument to top expr
					Ibuffer[Ip+tc_left] := Ip-TCI;
					Ibuffer[Ip+tc_right] := %1;
					! replace arg with complete
					! unary operation
					args[ap-1] := Ip;
				end
				else do
					Ip0 := %1;
				end
			end
			Ip := Ip+TCI;
		end
		else if (Ip0 \= %1) do
			! Current instr is not part of an expression:
			! Run the optimizations on the generated tree,
			! if any.
			if (ap = 2) commref(args);
			for (j=0, ap) do
				reorder(args[j]);
				writeback(args[j]);
			end
			! Ip0 < current position after update?
			! Fill with NOPs.
			if (Ip0 < i) do
				while (Ip0 < i) do
					Code::Ip0 := tcode.IGLUE;
					Ip0 := Ip0+1;
				end
				optdone := 1;
			end
			Ip0 := %1;	! delete tree
		end
		i := next(i);
	end
	if (optdone) condense();
end

! Redirect jumps to jumps and eliminate zero-length jumps
jumpredir() do var i, j, k, jrm;
	for (i=0, NLABELS) Labels[i] := %1;
	jrm := 0;	! jumps removed?
	i := 0;
	! remove zero-length jumps
	while (i < Ctop) do
		if (Code::i = tcode.ICLAB) do
			j := getw(i+1) - Lbase;
			if (j >= NLABELS) fatal("too many labels", 0);
			! make label (i+2+1 = code following CLAB)
			Labels[j] := i+2+1;
			! JUMP x CLAB x --> CLAB x
			if (	i >= 2+1 /\ Code::(i-2-1) = tcode.IJUMP /\
				getw(i-2) = j+Lbase
			) do
				nop3(i-2-1);
				jrm := 1;
			end
		end
		i := next(i);
	end
	i := 0;
	! jump-to-jump redirection
	while (i < Ctop) do
		if (Code::i = tcode.IJUMP) do
			j := i;		! remember beginning of jump chain
			while (1) do
				if (Code::j \= tcode.IJUMP) leave;
				k := getw(j+1) - Lbase;
				if (k < 0 \/ k >= NLABELS \/ Labels[k] = %1)
					leave;
				j := Labels[k];
				if (i = j) leave;	! cycle detection
			end
			if (i \= j) putw(i+1, k+Lbase);
		end
		i := next(i);
	end
	if (jrm) condense();
end

writeproc() if (Ctop /\ t.write(T3X.SYSOUT, Code, Ctop) \= Ctop)
		fatal("file write error", 0);

do var eof;
	init();
	eof := 0;
	while (\eof) do
		eof := readproc();
		foldexpr();
		foldcond();
		condense();
		treeopts();
		jumpredir();
		writeproc();
	end
end
