/*
 * TXX -- A Virtual Tcode Machine
 * Nils M Holm 1997-2003,2019
 * See the file LICENSE for conditions of use.
 *
 * SYNTAX: TXX PROGRAM [OPTIONS]
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>

#ifdef __TURBOC__
 #include <alloc.h>
 #include <io.h>
#endif

#ifndef __TURBOC__
 #include <unistd.h>
#endif

#define MAIN
#include "txx.h"
#include "tcode.h"
#include "rtcode.h"

void	ifail();

#ifdef dos
 #define TEXTSIZE	63488U
 #define DATASIZE	TEXTSIZE
#else
 #ifdef SMALL
  #define TEXTSIZE	29696
  #define DATASIZE	29696
 #else
  #define TEXTSIZE	65536L
  #define DATASIZE	TEXTSIZE
 #endif
#endif

#ifdef SMALL
 #define NLABELS	1536
 #define NMARKS		128
#else
 #define NLABELS	4096
 #define NMARKS		512
#endif

#define LDBUFLEN	1024

struct mark {
	int	l, t, d;
};

#define MAXINFS		256

struct inf {
	char	*name;
	uint	cmap;
	int	narg;
	xcell	(*proc)();
};

/* shared variables are defined in 'txx.h' */
uint		Textsize;
uint		Datasize;
uchar		*Text;
cell		*Wdata;
ucell		*Uwdata;
ucell		Etext, Edata;
int		Tags[NLABELS];
struct mark	Marks[NMARKS];
int		Mt;
uint		Dptr, Tptr;
uint		Op;
cell		A1, A2;
ucell		U1;
ucell		Ip, Fp;
ucell		Self;
int		Execmax;

uchar		Ldbuf[LDBUFLEN];
int		Lfd, Lp, Lz;

extern struct infproc	InfT3X[];
#ifdef EXT_SYSTEM
extern struct infproc	InfSystem[];
#endif
#ifdef EXT_TTYCTL
extern struct infproc	InfTTYCtl[];
#endif
#ifdef EXT_XMEM
extern struct infproc	InfXMem[];
#endif

struct inf	Infs[MAXINFS], *Lastinf = NULL;

struct infproc	*Interfaces[] = {
	InfT3X,
#ifdef EXT_SYSTEM
	InfSystem,
#endif
#ifdef EXT_TTYCTL
	InfTTYCtl,
#endif
#ifdef EXT_XMEM
	InfXMem,
#endif
	NULL };

#ifdef __TURBOC__
void fail(char *m, char *n);
#endif

void fail(m, n)
char	*m, *n;
{
	if (n)
		fprintf(stderr, "TXX core fault: %s: %s\n", m, n);
	else
		fprintf(stderr, "TXX core fault: %s\n", m);
	fprintf(stderr, "IP=%u, SP=%u, FP=%u, RR=%d, SELF=%u\n"
			"Etext=%u, Edata=%u\n",
		Ip, Sp<<1, Fp<<1, Acc, Self, Etext, Edata);
	if (Op) {
		fprintf(stderr, "Code: ");
		if (Op & 128) {
			if (Op == IINCL || Op == IINCG || Op == IINCI)
				fprintf(stderr, "%s %d (%u) %d\n",
					Tcode_Opcodes[Op&127], A1, U1, A2);
			else
				fprintf(stderr, "%s %d (%u)\n",
					Tcode_Opcodes[Op&127], A1, U1);
		}
		else {
			fprintf(stderr, "%s\n", Tcode_Opcodes[Op]);
		}
	}
	exit(2);
}

#ifdef SMALL

void initmem(prog)
char	*prog;
{
	struct stat	sb;

	Textsize = TEXTSIZE;
	Datasize = DATASIZE;
	if (prog && stat(prog, &sb) >= 0) {
		Textsize = sb.st_size;
		Datasize = TEXTSIZE+DATASIZE - Textsize;
	}
        Text = (uchar *) malloc(Textsize);
	Data = (uchar *) malloc(Datasize);
}

#else

void initmem(prog)
char	*prog;
{
        Text = (uchar *) malloc(TEXTSIZE);
	Data = (uchar *) malloc(DATASIZE);
	Textsize = TEXTSIZE;
	Datasize = DATASIZE;
}

#endif

void init(prog)
char	*prog;
{
	int	i;

	initmem(prog);
	if (!Text || !Data) fail("insufficient memory", 0);
	memset(Data, 0, Datasize);
        Wdata = (cell *) Data;
	Stack = Wdata;
	Uwdata = (ucell *) Data;
	Ustack = Uwdata;
	Sp = Fp = Datasize/2;
	Etext = 0;
	Edata = 0;
	for (i=0; i<NLABELS; i++) Tags[i] = -1;
	Mt = 0;
	Tptr = 0;
	Dptr = 0;
	A1 = 0;
	A2 = 0;
	Op = 0;
	Self = 0;
	Execmax = sizeof(Primitives) / sizeof(Primitives[0]);
	libinit();
}

void addlabel(op, id)
int	op, id;
{
	int	i;
	uint	a, o;
	uint	here;

	if (id >= NLABELS) fail("label too big", 0);
	if (Tags[id] != -1) fail("duplicate label", 0);
	Tags[id] = op == ICLAB? Tptr: Dptr;
	here = op == IDLAB? Dptr: Tptr;
	for (i=0; i<Mt; i++) if (Marks[i].l == id) break;
	if (i >= Mt) return;
	a = Marks[i].t;
	while (a) {
		o = a;
		a = (Text[a+1] << 8) | Text[a];
		Text[o] = here & 0xff;
		Text[o+1] = here >> 8;
	}
	Marks[i].t = 0;
	a = Marks[i].d;
	while (a) {
		o = a;
		a = (Data[a+1] << 8) | Data[a];
		Data[o] = here & 0xff;
		Data[o+1] = here >> 8;
	}
	Marks[i].d = 0;
}

int addmark(id, dref)
int	id, dref;
{
	int	i, old;

	for (i=0; i<Mt; i++) {
		if (Marks[i].l == id) {
			if (dref) {
				old = Marks[i].d;
				Marks[i].d = Dptr;
			}
			else {
				old = Marks[i].t;
				Marks[i].t = Tptr;
			}
			return old;
		}
	}
	for (i=0; i<Mt; i++) {
		if (Marks[i].d == 0 && Marks[i].t == 0)
			break;
	}
	if (i >= NMARKS) fail("too many marks", 0);
	Marks[i].l = id;
	if (dref) {
		Marks[i].d = Dptr;
		Marks[i].t = 0;
	}
	else {
		Marks[i].t = Tptr;
		Marks[i].d = 0;
	}
	if (i == Mt) Mt++;
	return 0;
}

int findlab(id, dref)
int	id, dref;
{
	if (Tags[id] != -1) return Tags[id];
	return addmark(id, dref);
}

uchar rdc() {
	if (Lp >= Lz) {
		Lz = read(Lfd, Ldbuf, LDBUFLEN);
		Lp = 0;
	}
	return Ldbuf[Lp++];
}

uint rdop() {
	uint	n = rdc();
	return (rdc() << 8) | n;
}

void skip(n)
int	n;
{
	int	i;

	for (i=0; i<n; i++) rdc();
}

void copy(n, p)
int	n;
uchar	*p;
{
	int	i;

	for (i=0; i<n; i++) p[i] = rdc();
}

void dgrow(n)
int	n;
{
	if ((Dptr += n) > Datasize) fail("data segment overflow", 0);
}

void tgrow(n)
int	n;
{
	if ((Tptr += n) > Textsize) fail("text segment overflow", 0);
}

void addinf(id, name, len)
int	id;
char	*name;
int	len;
{
	int	i, j;

	name[len] = 0;
	if (id >= MAXINFS)
		fail("interface ID out of range", name);
	if (Infs[id].name) {
		if (strcmp(Infs[id].name, name))
			fail("internal error: interface collision", name);
		return;
	}
	if ((Infs[id].name = (char *) malloc(len+1)) == NULL)
		fail("insufficient memory", 0);
	memcpy(Infs[id].name, name, len+1);
	for (j=0; Interfaces[j]; j++) {
		for (i=0; Interfaces[j][i].name; i++) {
			if (!strcmp(Interfaces[j][i].name, name)) {
				Infs[id].proc = Interfaces[j][i].proc;
				Infs[id].cmap = 0;
				Infs[id].narg = 0;
				Lastinf = &Infs[id];
				return;
			}
		}
	}
	fail("unsupported interface", name);
}

int resolve() {
	ucell	op, arg, arg2, a2f;
	uint	b;
	char	buf[257];

	op = rdc();
	if (op == '#') {
		while (Ldbuf[Lp++] != '\n')
			if (Lp >= LDBUFLEN-1)
				fail("#! header too long", 0);
	}
	else {
		Lp--;
	}
	if (Ldbuf[Lp] != IINIT)
		fail("bad input program (magic match failed)", 0);
	for (;;) {
		op = rdc();
		if (Lz < 1) break;
		if (op & 128) arg = rdop();
		a2f = 0;
		if (op==IINCG || op==IINCL || op==IINCI || op==IINIT) {
			arg2 = rdop();
			a2f = 1;
		}
		switch (op) {
		case ICLAB:
		case IDLAB:
			addlabel(op, arg);
			op = 0;
			break;
		case IVEC:
			dgrow(arg? (arg<<1): 2);
			op = 0;
			break;
		case IDATA:
			Wdata[Dptr>>1] = arg;
			dgrow(2);
			op = 0;
			break;
		case ICREF:
		case IDREF:
			Wdata[Dptr>>1] = findlab(arg, 1);
			dgrow(2);
			op = 0;
			break;
		case ISTR:
			b = Dptr;
			dgrow(arg);
			copy(arg, &Data[b]);
			do { Data[Dptr++] = 0; } while (Dptr & 1);
			op = 0;
			break;
		case IINCG: case ILDG: case ILDGV: case ISAVG:
		case ILDLAB: case ICALL: case IBRF: case IBRT:
		case INBRF: case INBRT: case IJUMP: case IUNEXT:
		case IDNEXT:
			Text[Tptr] = op;
			tgrow(1);
			arg = findlab(arg, 0);
			break;
		case IPUB:
			arg = rdop();
			skip(arg);
			op = 0;
			break;
		case ILSYM:
		case IGSYM:
		case IISYM:
			arg = rdop();
			skip(arg);
			op = 0;
			break;
		case IEXT:
			arg = rdop();
			copy(arg, Text);
			Text[arg] = 0;
			fail("unresolved external reference", (char *) Text);
			break;
		case IIREF:
			arg = rdop();
			copy(arg, Text);
			Text[arg] = 0;
			fail("unresolved interface reference", (char *) Text);
			break;
		case IHINT:
			op = 0;
			break;
		case IINIT:
			if (arg != 7) fail("unsupported tcode version", NULL);
			op = 0;
			break;
		case ILIB:
			skip(arg);
			op = 0;
			break;
		case IIPROC:
			arg2 = rdop();
			copy(arg2, (uchar *) buf);
			addinf(arg, buf, arg2);
			op = 0;
			break;
		case ICMAP:
			arg2 = rdop();
			if (!Lastinf) fail("CMAP without IPROC", 0);
			Lastinf->narg = arg;
			Lastinf->cmap = arg2;
			op = 0;
			break;
		default:
			Text[Tptr] = op;
			tgrow(1);
		}
		if (op & 128) {
			Text[Tptr] = arg & 0xff;
			Text[Tptr+1] = arg >> 8;
			tgrow(2);
			if (a2f) {
				Text[Tptr] = arg2 & 0xff;
				Text[Tptr+1] = arg2 >> 8;
				tgrow(2);
			}
		}
	}
	Etext = Tptr;
	Edata = Dptr;
	return 0;
}

void load(name)
char	*name;
{
	if ((Lfd = open(name, 0)) < 0) fail("cannot open program", name);
	Lp = Lz = 0;
	resolve();
	close(Lfd);
}

void primitive(id)
int	id;
{
	if (id >= Execmax-1) fail("undefined primitive", 0);
	(*Primitives[id])();
}

void infcall(id)
uint	id;
{
#ifdef ANSIPROTO
	xcell	(*p0)(),
		(*p1)(xcell),
		(*p2)(xcell, xcell),
		(*p3)(xcell, xcell, xcell),
		(*p4)(xcell, xcell, xcell, xcell),
		(*p5)(xcell, xcell, xcell, xcell, xcell),
		(*p6)(xcell, xcell, xcell, xcell, xcell, xcell),
		(*p7)(xcell, xcell, xcell, xcell, xcell, xcell, xcell),
		(*p8)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell),
		(*p9)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell),
		(*p10)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell, xcell),
		(*p11)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell, xcell, xcell),
		(*p12)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell, xcell, xcell, xcell),
		(*p13)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell, xcell, xcell, xcell, xcell),
		(*p14)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell, xcell, xcell, xcell, xcell, xcell),
		(*p15)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell),
		(*p16)(xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell, xcell, xcell, xcell, xcell, xcell,
			xcell, xcell);
#else
	xcell	(*p0)(), (*p1)(), (*p2)(), (*p3)(), (*p4)(), (*p5)(),
		(*p6)(), (*p7)(), (*p8)(), (*p9)(), (*p10)(), (*p11)(),
		(*p12)(), (*p13)(), (*p14)(), (*p15)(), (*p16)();
#endif

	#define	P0
	#define	P1	P[0]
	#define	P2	P[1],P[0]
	#define	P3	P[2],P[1],P[0]
	#define	P4	P[3],P[2],P[1],P[0]
	#define	P5	P[4],P[3],P[2],P[1],P[0]
	#define	P6	P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P7	P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P8	P[7],P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P9	P[8],P[7],P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P10	P[9],P[8],P[7],P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P11	P[10],P[9],P[8],P[7],P[6],P[5],P[4],P[3],P[2], \
			P[1],P[0]
	#define	P12	P[11],P[10],P[9],P[8],P[7],P[6],P[5],P[4],P[3], \
			P[2],P[1],P[0]
	#define	P13	P[12],P[11],P[10],P[9],P[8],P[7],P[6],P[5],P[4], \
			P[3],P[2],P[1],P[0]
	#define	P14	P[13],P[12],P[11],P[10],P[9],P[8],P[7],P[6],P[5], \
			P[4],P[3],P[2],P[1],P[0]
	#define	P15	P[14],P[13],P[12],P[11],P[10],P[9],P[8],P[7],P[6], \
			P[5],P[4],P[3],P[2],P[1],P[0]
	#define	P16	P[15],P[14],P[13],P[12],P[11],P[10],P[9],P[8],P[7], \
			P[6],P[5],P[4],P[3],P[2],P[1],P[0]
	xcell	P[16];
	int	i, n;
	uint	b, m;

	if (id >= MAXINFS || Infs[id].name == NULL)
		fail("bad interface call", 0);
	n = Infs[id].narg;
	m = Infs[id].cmap;
	for (b=1,i=0; i<n; i++) {
		if (m & b) {
			P[i] = Ustack[Sp+n-i];
			/* The next statement relies upon
			 * sizeof(xcell) == sizeof(char*).
			 * It may cause a warning.
			*/
			P[i] = (xcell) &Data[P[i]];
		}
		else {
			P[i] = Stack[Sp+n-i];
		}
		b <<= 1;
	}
	switch(n) {
	case  0: p0 = Infs[id].proc; Acc = (*p0)(P0); break;
	case  1: p1 = Infs[id].proc; Acc = (*p1)(P1); break;
	case  2: p2 = Infs[id].proc; Acc = (*p2)(P2); break;
	case  3: p3 = Infs[id].proc; Acc = (*p3)(P3); break;
	case  4: p4 = Infs[id].proc; Acc = (*p4)(P4); break;
	case  5: p5 = Infs[id].proc; Acc = (*p5)(P5); break;
	case  6: p6 = Infs[id].proc; Acc = (*p6)(P6); break;
	case  7: p7 = Infs[id].proc; Acc = (*p7)(P7); break;
	case  8: p8 = Infs[id].proc; Acc = (*p8)(P8); break;
	case  9: p9 = Infs[id].proc; Acc = (*p9)(P9); break;
	case 10: p10 = Infs[id].proc; Acc = (*p10)(P10); break;
	case 11: p11 = Infs[id].proc; Acc = (*p11)(P11); break;
	case 12: p12 = Infs[id].proc; Acc = (*p12)(P12); break;
	case 13: p13 = Infs[id].proc; Acc = (*p13)(P13); break;
	case 14: p14 = Infs[id].proc; Acc = (*p14)(P14); break;
	case 15: p15 = Infs[id].proc; Acc = (*p15)(P15); break;
	case 16: p16 = Infs[id].proc; Acc = (*p16)(P16); break;
	default: fail("too many arguments in ICALL", Infs[id].name);
	}
}

void divz() { fail("divide by zero", 0); }

void ifail() { Op = 0; fail("illegal opcode", 0); }

void iadd() { Stack[Sp+1] += Stack[Sp]; Sp++; }
void iband() { Stack[Sp+1] &= Stack[Sp]; Sp++; }
void ibnot() { Stack[Sp] = ~Stack[Sp]; }
void ibor() { Stack[Sp+1] |= Stack[Sp]; Sp++; }
void ibrf() { if (!Stack[Sp++]) Ip = U1; }
void ibrt() { if (Stack[Sp++]) Ip = U1; }
void ibshl() { Stack[Sp+1] <<= Stack[Sp]; Sp++; }
void ibshr() { Ustack[Sp+1] >>= Ustack[Sp]; Sp++; }
void ibxor() { Stack[Sp+1] ^= Stack[Sp]; Sp++; }
void icall() { Stack[--Sp] = Ip; Ip = U1; }
void icalr() { ucell t; t = Stack[Sp]; Stack[Sp] = Ip; Ip = t; }
void icalx() { fail("call to external procedure", 0); }
void iclab() { ifail(); }
void iclean() { Sp += A1; Stack[--Sp] = Acc; }
void icmap() { ifail(); }
void icref() { ifail(); }
void idata() { ifail(); }
void ideref() { Stack[Sp+1] += (Stack[Sp]<<1); Sp++;
	Stack[Sp] = Wdata[Ustack[Sp]>>1]; }
void idiv() { if (!Stack[Sp]) divz(); Stack[Sp+1] /= Stack[Sp]; Sp++; }
void idlab() { ifail(); }
void idnext() { if (Stack[Sp+1] <= Stack[Sp]) Ip = U1; Sp+=2; }
void idref() { ifail(); }
void idrefb() { Stack[Sp+1] += Stack[Sp]; Sp++; Stack[Sp] = Data[Ustack[Sp]]; }
void idup() { Stack[Sp-1] = Stack[Sp]; Sp--; }
void iend() { Fp = Stack[Sp++]; Ip = Stack[Sp++]; }
void iendm() { Self = Stack[Sp++]; Fp = Stack[Sp++]; Ip = Stack[Sp++]; }
void iequ() { Stack[Sp+1] = Stack[Sp+1] == Stack[Sp]? -1: 0; Sp++; }
void iext() { ifail(); }
void iglue() { }
void igrtr() { Stack[Sp+1] = Stack[Sp+1] > Stack[Sp]? -1: 0; Sp++; }
void igsym() { ifail(); }
void igteq() { Stack[Sp+1] = Stack[Sp+1] >= Stack[Sp]? -1: 0; Sp++; }
void ihalt() { exit(A1); }
void ihdr() { Stack[--Sp] = Fp; Fp = Sp; }
void ihint() { ifail(); }
void iicall() { infcall(U1); }
void iicalx() { fail("call to external interface", 0); }
void iincg() { A2 = (Text[Ip+1] << 8) | Text[Ip]; Ip += 2;
	Wdata[U1>>1] += A2; }
void iinci() { A2 = (Text[Ip+1] << 8) | Text[Ip]; Ip += 2;
	Wdata[(Self>>1)+U1] += A2; }
void iincl() { A2 = (Text[Ip+1] << 8) | Text[Ip]; Ip += 2;
	Stack[Fp-A1] += A2; }
void iinit() { ifail(); }
void iiproc() { ifail(); }
void iiref() { ifail(); }
void iisym() { ifail(); }
void ijump() { Ip = U1; }
void ildg() { Stack[--Sp] = Wdata[U1>>1]; }
void ildgv() { Stack[--Sp] = U1; }
void ildi() { Stack[--Sp] = Wdata[(Self>>1) + U1]; }
void ildiv() { Stack[--Sp] = Self + (U1<<1); }
void ildl() { Stack[--Sp] = Stack[Fp-A1]; }
void ildlab() { Stack[--Sp] = U1; }
void ildlv() { Stack[--Sp] = (Fp-A1)<<1; }
void iless() { Stack[Sp+1] = Stack[Sp+1] < Stack[Sp]? -1: 0; Sp++; }
void ilib() { ifail(); }
void iline() { }
void ilnot() { Stack[Sp] = Stack[Sp]? 0: -1; }
void ilsym() { ifail(); }
void ilteq() { Stack[Sp+1] = Stack[Sp+1] <= Stack[Sp]? -1: 0; Sp++; }
void imhdr() { Stack[--Sp] = Fp; Fp = Sp;
	Stack[--Sp] = Self; Self = Stack[Sp+3]; }
void imod() { if (!Stack[Sp]) divz(); Ustack[Sp+1] %= Ustack[Sp]; Sp++; }
void imul() { Stack[Sp+1] *= Stack[Sp]; Sp++; }
void inbrf() { if (!Stack[Sp]) Ip = U1; }
void inbrt() { if (Stack[Sp]) Ip = U1; }
void ineg() { Stack[Sp] = -Stack[Sp]; }
void inequ() { Stack[Sp+1] = Stack[Sp+1] != Stack[Sp]? -1: 0; Sp++; }
void inorm() { Stack[Sp+1] += (Stack[Sp]<<1); Sp++; }
void inormb() { Stack[Sp+1] += Stack[Sp]; Sp++; }
void inum() { Stack[--Sp] = A1; }
void ipop() { Acc = Stack[Sp++]; }
void ipstr() { ifail(); }
void ipub() { ifail(); }
void isavg() { Wdata[U1>>1] = Stack[Sp++]; }
void isavi() { Wdata[(Self>>1)+U1] = Stack[Sp++]; }
void isavl() { Stack[Fp-A1] = Stack[Sp++]; }
void iself() { Stack[--Sp] = Self; }
void istack() { Sp -= A1; if ((Sp<<1) < Edata) fail("stack overflow", 0); }
void istorb() { Data[Ustack[Sp+1]] = Ustack[Sp]; Sp+=2; }
void istore() { Wdata[Ustack[Sp+1]>>1] = Stack[Sp]; Sp+=2; }
void istr() { ifail(); }
void isub() { Stack[Sp+1] -= Stack[Sp]; Sp++; }
void iswap() { ucell t; t = Stack[Sp+1]; Stack[Sp+1] = Stack[Sp];
	Stack[Sp] = t; }
void isys() { primitive(U1); }
void iudiv() { if (!Stack[Sp]) divz(); Ustack[Sp+1] /= Ustack[Sp]; Sp++; }
void iugrtr() { Ustack[Sp+1] = Ustack[Sp+1] > Ustack[Sp]? -1: 0; Sp++; }
void iugteq() { Ustack[Sp+1] = Ustack[Sp+1] >= Ustack[Sp]? -1: 0; Sp++; }
void iuless() { Ustack[Sp+1] = Ustack[Sp+1] < Ustack[Sp]? -1: 0; Sp++; }
void iulteq() { Ustack[Sp+1] = Ustack[Sp+1] <= Ustack[Sp]? -1: 0; Sp++; }
void iumul() { Ustack[Sp+1] *= Ustack[Sp]; Sp++; }
void iunext() { if (Stack[Sp+1] >= Stack[Sp]) Ip = U1; Sp+=2; }
void ivec() { ifail(); }

void run() {
	uint	o;

	for (Ip = 0;;) {
		Op = Text[Ip++];
		if (Op & 128) {
			A1 = (Text[Ip+1] << 8) | Text[Ip];
			Ip += 2;
		}
		o = Op & ~128;
		if (o >= IENDOFSET) ifail();
		U1 = A1;
		(*Tcode_instructions[o])();
	}
}

void version() {
	printf("TXX -- Virtual Tcode Machine, Version 5.5 (Tcode7)\n");
	printf("Nils M Holm, 1997-2019\n");
#ifdef SMALL
	printf("BPW = %d, Text+Data = %dKB combined, Labels = %d\n",
		sizeof(cell), Textsize/1024 + Datasize/1024,
		NLABELS);
#else
	printf("BPW = %d, Text = %dKB, Data = %dKB, Labels = %d\n",
		sizeof(cell), Textsize/1024,
		Datasize/1024, NLABELS);
#endif
	printf("Extensions = T3X");
#ifdef EXT_SYSTEM
	printf(", System");
#endif
#ifdef EXT_TTYCTL
	printf(", TTYCtl");
#endif
#ifdef EXT_XMEM
	printf(", XMem");
#endif
	putchar('\n');
	exit(0);
}

int main(argc, argv)
int	argc;
char	**argv;
{
	Argcount = argc-1;
	Argvec = argv+1;
	init(argv[1]);
	if (argc < 2) version();
	load(argv[1]);
	run();
	return 0;
}
