/*
 * T3X -- Common Runtime Support Routines for T3X R7.
 * Nils M Holm, 1999,2000,2002,2003,2019.
 * See the file LICENSE for conditions of use.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#ifdef __TURBOC__
 #include <io.h>
#else
 #include <unistd.h>
#endif

#include "../compiler/txx.h"

#ifdef __TURBOC__
xcell t3x_bpw S0();
xcell t3x_memcopy S3(char *d, char *s, xcell n);
xcell t3x_memcomp S3(char *s1, char *s2, xcell n);
xcell t3x_memfill S3(char *p, xcell c, xcell n);
xcell t3x_memscan S3(char *p, xcell c, xcell l);
xcell t3x_getarg S3(xcell i, char *b, xcell n);
xcell t3x_getenv S3(char *vn, char *b, xcell n);
xcell t3x_open S2(char *file, xcell mode);
xcell t3x_seek S3(xcell fd, xucell n, xcell o);
xcell t3x_close S1(xcell fd);
xcell t3x_read S3(xcell fd, char *b, xcell n);
xcell t3x_write S3(xcell fd, char *b, xcell n);
uchar *t3x_newline S1(uchar *s);
xcell t3x_remove S1(char *s);
xcell t3x_rename S2(char *so, char *sn);
xcell t3x_cvalist S4(xcell n, xucell bmap, xcell *ilist, xcell *olist);
#endif

void libinit() {
#ifdef dos
	setmode(0, O_BINARY);
	setmode(1, O_BINARY);
#endif
}

#ifdef LIBRARY

int	Argcount;
char	**Argvec;

void	_TXSTART_();

int main(argc, argv)
int	argc;
char	**argv;
{
	Argcount = argc;
	Argvec = argv;
	libinit();
	_TXSTART_();
	return 0;
}

#endif

xcell t3x_bpw S0()
SELF
{
	return BPW;
}

xcell t3x_memcopy S3(d, s, n)
SELF
xcell	n;
char	*d, *s;
{
	memmove(d, s, n);
	return 0;
}

xcell t3x_memcomp S3(s1, s2, n)
SELF
xcell	n;
char	*s2, *s1;
{
	return memcmp(s1, s2, n);
}

xcell t3x_memfill S3(p, c, n)
SELF
xcell	n, c;
char	*p;
{
	memset(p, c, n);
	return 0;
}

xcell t3x_memscan S3(p, c, l)
SELF
xcell	l, c;
char	*p;
{
	int	i;

	for (i=0; i<l; i++)
		if (p[i] == c) return i;
	return -1;
}

xcell t3x_getarg S3(i, b, n)
SELF
xcell	n;
char	*b;
xcell	i;
{
	int	k;

	if (i >= Argcount) {
		b[0] = 0;
		return -1;
	}
	k = strlen(Argvec[i]);
	k = k>=n? n-1: k;
	memcpy(b, Argvec[i], k);
	b[k] = 0;
	return k;
}

xcell t3x_getenv S3(vn, b, n)
SELF
xcell	n;
char	*b;
char	*vn;
{
	char	*vv;
	int	k;

	vv = getenv(vn);
	if (vv == NULL) {
		b[0] = 0;
		return -1;
	}
	k = strlen(vv);
	k = k>=n? n-1: k;
	memcpy(b, vv, k);
	b[k] = 0;
	return k;
}

xcell t3x_open S2(file, mode)
SELF
xcell	mode;
char	*file;
{
	int	fd;

	if (mode == 1) {
		fd = creat(file, 0666);
	}
	else if (mode == 0 || mode == 2 || mode == 3) {
		fd = open(file, mode==3? O_WRONLY: mode==2? O_RDWR:
			O_RDONLY);
	}
	else {
		return -1;
	}
	if (fd > 0 && mode == 3) lseek(fd, 0L, 2);
	return fd;
}

#define T_SEEK_SET	0
#define T_SEEK_FWD	1
#define T_SEEK_END	2
#define T_SEEK_BCK	3

xcell t3x_seek S3(fd, n, o)
SELF
xcell	o;
xucell	n;
xcell	fd;
{
	n &= 0xffff;
	if (o == T_SEEK_SET)
		return lseek(fd, (off_t) n * 1L, SEEK_SET) == -1? -1: 0;
	else if (o == T_SEEK_END)
		return (int) lseek(fd, (off_t) n * -1L, SEEK_END) == -1? -1: 0;
	else if (o == T_SEEK_FWD)
		return (int) lseek(fd, (off_t) n * 1L, SEEK_CUR) == -1? -1: 0;
	else if (o == T_SEEK_BCK)
		return (int) lseek(fd, (off_t) n * -1L, SEEK_CUR) == -1? -1: 0;
	else
		return -1;
}

xcell t3x_close S1(fd)
SELF
xcell	fd;
{
	return close(fd);
}

xcell t3x_read S3(fd, b, n)
SELF
xcell	n;
char	*b;
xcell	fd;
{
	return read(fd, b, n);
}

xcell t3x_write S3(fd, b, n)
SELF
xcell	n;
char	*b;
xcell	fd;
{
	return write(fd, b, n);
}

uchar *t3x_newline S1(s)
SELF
uchar	*s;
{
#ifdef NL_IS_LF
	s[0] = '\n';
	s[1] = 0;
#endif
#ifdef NL_IS_CRLF
	s[0] = '\r';
	s[1] = '\n';
	s[2] = 0;
#endif
#ifdef LIBRARY
	return s;
#else
	return (uchar *) (s - Data);
#endif
}

xcell t3x_remove S1(s)
SELF
char	*s;
{
	return unlink(s);
}

xcell t3x_rename S2(so, sn)
SELF
char	*sn, *so;
{
	if (close(open(sn, 0)) >= 0) return -1;
	return rename(so, sn);
}

#ifdef HOST16

xcell t3x_cvalist S4(n, bmap, ilist, olist)
SELF
xcell	*olist;
xcell	*ilist;
xucell	bmap;
xcell	n;
{
	bmap = bmap;	/*USED*/
	if (n<0)
		memcpy(ilist, olist, (-n)*sizeof(int));
	else
		memcpy(olist, ilist, n*sizeof(int));
	return 0;
}

#endif /*HOST16*/

#ifndef HOST16

#ifdef LIBRARY

xcell t3x_cvalist S4(n, bmap, ilist, olist)
SELF
void	**olist;
void	*ilist;
xucell	bmap;
xcell	n;
{
	if (n<0)
		memcpy(ilist, olist, (-n)*sizeof(void*));
	else
		memcpy(olist, ilist, n*sizeof(void*));
	return 0;
}

#else	/* !LIBRARY */

xcell t3x_cvalist S4(n, bmap, ilist, olist)
SELF
void	**olist;
ushort	*ilist;
xucell	bmap;
xcell	n;
{
	int	i;
	ulong	p, o;

	/* Align olist at 8-byte word boundary */
	o = (ulong) olist;
	p = (o + 7) & ~7UL;
	olist = (void**) p;
	if (n < 0) {
		n = -n;
		for (i=0; i<n; i++)
			ilist[i] = (ushort) (WORDCAST olist[i]);
			/*YESIMEANIT!*/
	}
	else {
		for (i=0; i<n; i++) {
			if (bmap&1)
				olist[i] = &Data[ilist[i]];
			else
				olist[i] = (void *) (WORDCAST ilist[i]);
				/*JUSTDOIT!*/
			bmap>>=1;
		}
	}
	return (int)(p-o);
}

#endif	/* !LIBRARY */

#endif /*!HOST16*/

#ifndef LIBRARY

void X_sys()	{ Acc = 0; }
void X_bpw()	{ Acc = t3x_bpw(); }
void X_memcopy(){ Acc = t3x_memcopy(UARG(3,2), ARGP(3,1), ARGP(3,0)); }
void X_memcomp(){ Acc = t3x_memcomp(UARG(3,2), ARGP(3,1), ARGP(3,0)); }
void X_memfill(){ Acc = t3x_memfill(UARG(3,2), ARG(3,1), ARGP(3,0)); }
void X_memscan(){ Acc = t3x_memscan(UARG(3,2), ARG(3,1), ARGP(3,0)); }
void X_getarg()	{ Acc = t3x_getarg(ARG(3,2), ARGP(3,1), ARG(3,0)); }
void X_getenv()	{ Acc = t3x_getenv(ARG(3,2), ARGP(3,1), ARGP(3,0)); }
void X_open()	{ Acc = t3x_open(ARG(2,1), ARGP(2,0)); }
void X_seek()	{ Acc = t3x_seek(ARG(3,2), UARG(3,1), ARG(3,0)); }
void X_close()	{ Acc = t3x_close(ARG(1,0)); }
void X_read()	{ Acc = t3x_read(ARG(3,2), ARGP(3,1), ARG(3,0)); }
void X_write()	{ Acc = t3x_write(ARG(3,2), ARGP(3,1), ARG(3,0)); }
void X_newline(){ Acc = (xcell) t3x_newline(UARGP(1,0)); }
void X_remove()	{ Acc = t3x_remove(ARGP(1,0)); }
void X_rename()	{ Acc = t3x_rename(ARGP(2,1), ARGP(2,0)); }
#ifdef HOST16
 void X_cvalist(){ Acc = t3x_cvalist((xcell *) ARGP(4,3), (xcell *) ARGP(4,2),
			ARG(4,1), ARG(4,0)); }
#else
 #ifdef LIBRARY
  void X_cvalist(){ Acc = t3x_cvalist(ARGP(4,3), ARGP(4,2), ARG(4,1),
			ARG(4,0)); }
 #else
  void X_cvalist(){ Acc = t3x_cvalist((void **) ARGP(4,3),
			(ushort *) ARGP(4,2), ARG(4,1), ARG(4,0)); }
 #endif
#endif

struct infproc	InfT3X[] = {
	{ "t3x_bpw",		t3x_bpw		},
	{ "t3x_memcopy",	t3x_memcopy	},
	{ "t3x_memcomp",	t3x_memcomp	},
	{ "t3x_memfill",	t3x_memfill	},
	{ "t3x_memscan",	t3x_memscan	},
	{ "t3x_getarg",		t3x_getarg	},
	{ "t3x_getenv",		t3x_getenv	},
	{ "t3x_open",		t3x_open	},
	{ "t3x_read",		t3x_read	},
	{ "t3x_write",		t3x_write	},
	{ "t3x_newline",	(xcell(*)())
				t3x_newline	},
	{ "t3x_seek",		t3x_seek	},
	{ "t3x_close",		t3x_close	},
	{ "t3x_rename",		t3x_rename	},
	{ "t3x_remove",		t3x_remove	},
	{ "t3x_cvalist",	t3x_cvalist	},
	{ NULL,			NULL		}
};

#endif

#ifdef NEED_RENAME

rename(old, new)
char	*old, *new;
{
	int	r;

	if ((r = link(old, new)) < 0) return r;
	return unlink(old);
}

#endif
